from __future__ import annotations

import os
from pathlib import Path
import shutil
import stat
import subprocess

import pytest

FISH = shutil.which('fish')
COMPLETION_SCRIPT = Path(__file__).parents[2] / 'mycli' / 'resources' / 'completions' / 'fish' / 'mycli.fish'


@pytest.mark.skipif(FISH is None, reason='Fish is not installed')
def test_fish_completion_matches_mycli_completion_behavior(tmp_path: Path) -> None:
    assert FISH is not None
    executable = tmp_path / 'mycli'
    executable.write_text(
        '''#!/bin/sh
if [ "${_MYCLI_COMPLETE:-}" = 'fish_complete' ]; then
    printf 'click\n' >> "$MYCLI_LOG"
    if [ "$COMP_CWORD" = '--s' ]; then
        printf 'plain,--socket\tSocket file\nplain,--ssl-mode\tTLS mode\n'
    elif [ "$COMP_CWORD" = 'ba' ]; then
        printf 'plain,bash\n'
    fi
    exit 0
fi
if [ "$1" = '--list-dsn' ]; then
    printf 'list-dsn\n' >> "$MYCLI_LOG"
    printf 'prod\nstaging\n'
fi
''',
        encoding='utf-8',
    )
    executable.chmod(executable.stat().st_mode | stat.S_IXUSR)
    (tmp_path / 'mysql.sock').touch()
    (tmp_path / 'space socket.sock').touch()
    (tmp_path / 'batch.sql').touch()
    log_path = tmp_path / 'calls.log'
    environment = {
        **os.environ,
        'COMPLETION_SCRIPT': str(COMPLETION_SCRIPT),
        'MYCLI_LOG': str(log_path),
        'PATH': f'{tmp_path}{os.pathsep}{os.environ.get("PATH", "")}',
    }

    result = subprocess.run(
        [FISH, '--no-config', '-c', FISH_COMPLETION_HARNESS],
        check=True,
        capture_output=True,
        text=True,
        cwd=tmp_path,
        env=environment,
    )

    assert result.stdout.splitlines() == [
        'generic|--socket\tSocket file|--ssl-mode\tTLS mode',
        'completions|bash',
        'bare|prod',
        'dsn-separate|prod|staging',
        'dsn-long-separate|prod',
        'dsn-attached|-dprod',
        'dsn-long-attached|--dsn=prod',
        'database-separate|staging',
        'database-short-separate|staging',
        'database-short-attached|-Dprod',
        'database-attached|--database=prod',
        'after-host|prod',
        'host-value',
        'after-double-dash|staging',
        'after-positional',
        'clustered-option|prod',
        'socket-separate|mysql.sock',
        'socket-attached|--socket=space socket.sock',
        'socket-short-separate|space socket.sock',
        'socket-short-attached|-Smysql.sock',
        'checkpoint-separate|batch.sql',
        'checkpoint-attached|--checkpoint=batch.sql',
        'batch-separate|batch.sql',
        'batch-attached|--batch=batch.sql',
    ]
    assert log_path.read_text(encoding='utf-8').splitlines().count('list-dsn') == 12


FISH_COMPLETION_HARNESS = r'''
complete --erase --command mycli
source "$COMPLETION_SCRIPT"

function run_completion --argument-names label command
    printf '%s' "$label"
    for candidate in (complete -C "$command")
        printf '|%s' "$candidate"
    end
    printf '\n'
end

run_completion generic 'mycli --s'
run_completion completions 'mycli --completions ba'
run_completion bare 'mycli pro'
run_completion dsn-separate 'mycli -d '
run_completion dsn-long-separate 'mycli --dsn pro'
run_completion dsn-attached 'mycli -dpro'
run_completion dsn-long-attached 'mycli --dsn=pro'
run_completion database-separate 'mycli --database sta'
run_completion database-short-separate 'mycli -D sta'
run_completion database-short-attached 'mycli -Dpro'
run_completion database-attached 'mycli --database=pro'
run_completion after-host 'mycli --host db pro'
run_completion host-value 'mycli --host pro'
run_completion after-double-dash 'mycli -- sta'
run_completion after-positional 'mycli prod --host db other'
run_completion clustered-option 'mycli -vP 3306 pro'
run_completion socket-separate 'mycli --socket mysql'
run_completion socket-attached 'mycli --socket=space'
run_completion socket-short-separate 'mycli -S space'
run_completion socket-short-attached 'mycli -Smysql'
run_completion checkpoint-separate 'mycli --checkpoint batch'
run_completion checkpoint-attached 'mycli --checkpoint=bat'
run_completion batch-separate 'mycli --batch batch'
run_completion batch-attached 'mycli --batch=bat'
'''



HOME_URL = 'https://mycli.net'
REPO_URL = 'https://github.com/dbcli/mycli'
DOCS_URL = f'{HOME_URL}/docs'
ISSUES_URL = f'{REPO_URL}/issues'

DEFAULT_CHARSET = 'utf8mb4'
KNOWN_DSN_QUERY_PARAMS = {
    'character_set',
    'keepalive_ticks',
    'prompt',
    'socket',
    'ssh_jump',
    'ssl_ca',
    'ssl_capath',
    'ssl_cert',
    'ssl_cipher',
    'ssl_key',
    'ssl_mode',
    'ssl_verify_server_cert',
    'tls_version',
    'vault_address',
    'vault_mount',
    'vault_password_field',
    'vault_secret',
    'vault_username_field',
}
DEFAULT_DATABASE = 'mysql'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 3306
DEFAULT_USER = 'root'

TEST_DATABASE = 'mycli_test_db'

DEFAULT_WIDTH = 80
DEFAULT_HEIGHT = 25

# MySQL error codes not available in pymysql.constants.ER
ER_MUST_CHANGE_PASSWORD_LOGIN = 1862
ER_MUST_CHANGE_PASSWORD = 1820

EMPTY_PASSWORD_FLAG_SENTINEL = -1
DEFAULT_PROMPT = "\\t \\u@\\h:\\d> "



from __future__ import annotations

from importlib import resources

import mycli as mycli_package

COMPLETION_PATHS = {
    'bash': ('resources', 'completions', 'bash', 'mycli'),
    'zsh': ('resources', 'completions', 'zsh', '_mycli'),
    'fish': ('resources', 'completions', 'fish', 'mycli.fish'),
}


def main_completions(shell: str) -> None:
    completion_path = resources.files(mycli_package).joinpath(*COMPLETION_PATHS[shell])
    print(completion_path.read_text(), end='')



from collections.abc import Mapping
from types import SimpleNamespace
from typing import Any
from urllib.parse import parse_qsl, urlsplit

from mycli.constants import KNOWN_DSN_QUERY_PARAMS
from mycli.packages.special.dsn_aliases import INVALID_DSN_ALIAS_ERROR, DsnAliases, is_valid_dsn_alias


class DummyConfig(dict):
    def __init__(self, initial: Mapping[str, object] | None = None) -> None:
        super().__init__(initial or {})
        self.encoding: str | None = None
        self.write_calls = 0

    def write(self) -> None:
        self.write_calls += 1


def test_is_valid_dsn_alias_rejects_dash_prefix() -> None:
    assert is_valid_dsn_alias('prod') is True
    assert is_valid_dsn_alias('-prod') is False


def test_from_config_returns_instance_with_same_config() -> None:
    config = DummyConfig()

    aliases = DsnAliases.from_config(config)

    assert isinstance(aliases, DsnAliases)
    assert aliases.config is config


def test_from_config_retains_mycli_runtime() -> None:
    config = DummyConfig()
    mycli: Any = SimpleNamespace()

    aliases = DsnAliases.from_config(config, mycli)  # type: ignore[arg-type]

    assert aliases.mycli is mycli


def test_query_param_defaults_without_runtime_returns_empty_dict() -> None:
    aliases = DsnAliases(DummyConfig())

    assert aliases._query_param_defaults() == {}


def test_query_param_defaults_treats_invalid_boolean_as_false() -> None:
    config = DummyConfig({
        'connection': {'default_ssl_verify_server_cert': 'invalid'},
    })
    mycli = SimpleNamespace(default_keepalive_ticks=30, ssl_mode='auto')
    aliases = DsnAliases(config, mycli)  # type: ignore[arg-type]

    assert aliases._query_param_defaults()['ssl_verify_server_cert'] is False


def test_list_and_get_use_alias_dsn_section() -> None:
    config = DummyConfig({
        'alias_dsn': {
            'prod': 'mysql://prod/db',
            'staging': 'mysql://staging/db',
            '-hidden': 'mysql://hidden/db',
        },
    })
    aliases = DsnAliases(config)

    assert aliases.list() == ['prod', 'staging']
    assert aliases.get('prod') == 'mysql://prod/db'
    assert aliases.get('missing') is None
    assert aliases.get('-hidden') is None


def test_list_returns_empty_list_when_section_is_missing() -> None:
    aliases = DsnAliases(DummyConfig())

    assert aliases.list() == []


def test_save_creates_section_sets_encoding_and_writes_config() -> None:
    config = DummyConfig()
    aliases = DsnAliases(config)

    result = aliases.save('prod', 'mysql://prod/db')

    assert result == 'Saved: prod'
    assert config.encoding == 'utf-8'
    assert config == {'alias_dsn': {'prod': 'mysql://prod/db'}}
    assert config.write_calls == 1


def test_save_updates_existing_section_and_writes_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.save('staging', 'mysql://staging/db')

    assert result == 'Saved: staging'
    assert config.encoding == 'utf-8'
    assert config['alias_dsn'] == {
        'prod': 'mysql://prod/db',
        'staging': 'mysql://staging/db',
    }
    assert config.write_calls == 1


def test_save_rejects_dash_prefixed_alias_without_writing_config() -> None:
    config = DummyConfig()
    aliases = DsnAliases(config)

    result = aliases.save('-prod', 'mysql://prod/db')

    assert result == INVALID_DSN_ALIAS_ERROR
    assert config.encoding is None
    assert config == {}
    assert config.write_calls == 0


def test_delete_removes_existing_alias_and_writes_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.delete('prod')

    assert result == 'Deleted: prod'
    assert config['alias_dsn'] == {}
    assert config.write_calls == 1


def test_delete_rejects_dash_prefixed_alias_without_writing_config() -> None:
    config = DummyConfig({'alias_dsn': {'-prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.delete('-prod')

    assert result == INVALID_DSN_ALIAS_ERROR
    assert config['alias_dsn'] == {'-prod': 'mysql://prod/db'}
    assert config.write_calls == 0


def test_delete_returns_not_found_without_writing_config() -> None:
    config = DummyConfig({'alias_dsn': {'prod': 'mysql://prod/db'}})
    aliases = DsnAliases(config)

    result = aliases.delete('missing')

    assert result == 'Not Found: missing'
    assert config['alias_dsn'] == {'prod': 'mysql://prod/db'}
    assert config.write_calls == 0


def test_delete_returns_not_found_when_section_is_missing() -> None:
    config = DummyConfig()
    aliases = DsnAliases(config)

    result = aliases.delete('missing')

    assert result == 'Not Found: missing'
    assert config == {}
    assert config.write_calls == 0


def test_dsn_more_adds_non_default_runtime_parameters_in_sorted_order() -> None:
    config = DummyConfig({
        'main': {'prompt': 'configured> '},
        'connection': {
            'default_character_set': 'latin1',
            'default_keepalive_ticks': '30',
            'default_socket': '/default.sock',
            'default_ssl_mode': 'auto',
            'default_ssl_ca': '/default-ca.pem',
            'default_ssl_verify_server_cert': 'False',
        },
        'vault_beta': {
            'address': 'https://default-vault',
            'default_mount': 'kv',
            'default_password_field': 'password',
            'default_username_field': 'username',
        },
    })
    mycli = SimpleNamespace(
        default_keepalive_ticks=30,
        keepalive_ticks=45,
        prompt_format='runtime> ',
        ssl_mode='auto',
        sqlexecute=SimpleNamespace(
            character_set='utf8',
            ssl={
                'mode': 'on',
                'ca': '/runtime-ca.pem',
                'capath': '/runtime-ca',
                'cert': '/client-cert.pem',
                'cipher': 'AES256',
                'key': '/client-key.pem',
                'check_hostname': True,
                'tls_version': 'TLSv1.3',
            },
        ),
    )
    aliases = DsnAliases(config, mycli)  # type: ignore[arg-type]
    dsn = (
        'mysql://user@host/db?socket=%2Fruntime.sock&ssh_jump=bastion'
        '&vault_address=https%3A%2F%2Fruntime-vault&vault_mount=runtime-kv'
        '&vault_secret=database%2Fprod&vault_password_field=secret&vault_username_field=login'
    )

    more_dsn = aliases.dsn_more(dsn)

    parsed = urlsplit(more_dsn)
    assert (parsed.scheme, parsed.netloc, parsed.path) == ('mysql', 'user@host', '/db')
    more_params = parse_qsl(parsed.query)
    assert {key for key, _value in more_params} == KNOWN_DSN_QUERY_PARAMS
    assert more_params == [
        ('character_set', 'utf8'),
        ('keepalive_ticks', '45'),
        ('prompt', 'runtime> '),
        ('socket', '/runtime.sock'),
        ('ssh_jump', 'bastion'),
        ('ssl_ca', '/runtime-ca.pem'),
        ('ssl_capath', '/runtime-ca'),
        ('ssl_cert', '/client-cert.pem'),
        ('ssl_cipher', 'AES256'),
        ('ssl_key', '/client-key.pem'),
        ('ssl_mode', 'on'),
        ('ssl_verify_server_cert', 'true'),
        ('tls_version', 'TLSv1.3'),
        ('vault_address', 'https://runtime-vault'),
        ('vault_mount', 'runtime-kv'),
        ('vault_password_field', 'secret'),
        ('vault_secret', 'database/prod'),
        ('vault_username_field', 'login'),
    ]


def test_dsn_more_omits_empty_false_and_active_default_parameters() -> None:
    config = DummyConfig({
        'main': {'prompt': 'configured> '},
        'connection': {
            'default_character_set': 'latin1',
            'default_socket': '/default.sock',
            'default_ssl_ca': '/default-ca.pem',
            'default_ssl_mode': 'on',
            'default_ssl_verify_server_cert': 'True',
        },
        'vault_beta': {
            'address': 'https://default-vault',
            'default_mount': 'kv',
            'default_password_field': 'secret',
            'default_username_field': 'login',
        },
    })
    mycli = SimpleNamespace(
        default_keepalive_ticks=30,
        keepalive_ticks=30,
        prompt_format='configured> ',
        ssl_mode='on',
        sqlexecute=SimpleNamespace(
            character_set='latin1',
            ssl={'mode': 'on', 'ca': '/default-ca.pem', 'check_hostname': True},
        ),
    )
    aliases = DsnAliases(config, mycli)  # type: ignore[arg-type]
    dsn = (
        'mysql://user@host/db?socket=%2Fdefault.sock&vault_address=https%3A%2F%2Fdefault-vault'
        '&vault_mount=kv&vault_password_field=secret&vault_username_field=login&vault_secret='
    )

    assert aliases.dsn_more(dsn) == 'mysql://user@host/db'


def test_dsn_more_without_runtime_returns_original_dsn() -> None:
    aliases = DsnAliases(DummyConfig())

    assert aliases.dsn_more('mysql://user@host/db?socket=%2Ftmp%2Fmysql.sock') == ('mysql://user@host/db?socket=%2Ftmp%2Fmysql.sock')


def test_dsn_more_without_sql_executor_returns_original_dsn() -> None:
    mycli = SimpleNamespace(sqlexecute=None)
    aliases = DsnAliases(DummyConfig(), mycli)  # type: ignore[arg-type]
    dsn = 'mysql://user@host/db?socket=%2Ftmp%2Fmysql.sock'

    assert aliases.dsn_more(dsn) == dsn



from __future__ import annotations

import os
from pathlib import Path
import shutil
import stat
import subprocess

import pytest


def find_supported_bash() -> str | None:
    candidates = [
        os.environ.get('MYCLI_TEST_BASH'),
        '/opt/homebrew/bin/bash',
        '/opt/local/bin/bash',
        '/usr/local/bin/bash',
        shutil.which('bash5'),
        shutil.which('bash4'),
        shutil.which('bash'),
    ]
    for candidate in dict.fromkeys(candidates):
        if candidate is None or not Path(candidate).is_file():
            continue
        version_check = subprocess.run(
            # completions mostly work with bash 3 but some tests would fail to match
            [candidate, '-c', '(( BASH_VERSINFO[0] > 4 || (BASH_VERSINFO[0] == 4 && BASH_VERSINFO[1] >= 4) ))'],
            check=False,
        )
        if version_check.returncode == 0:
            return candidate
    return None


BASH = find_supported_bash()
COMPLETION_SCRIPT = Path(__file__).parents[2] / 'mycli' / 'resources' / 'completions' / 'bash' / 'mycli'


@pytest.mark.skipif(BASH is None, reason='Bash 4.4 or newer is not installed')
def test_bash_completion_matches_mycli_completion_behavior(tmp_path: Path) -> None:
    assert BASH is not None
    executable = tmp_path / 'mycli'
    executable.write_text(
        '''#!/bin/sh
if [ "${_MYCLI_COMPLETE:-}" = 'bash_complete' ]; then
    printf 'click\\n' >> "$MYCLI_LOG"
    if [ "$COMP_WORDS" = 'mycli --s' ]; then
        printf 'plain,--socket\\nplain,--ssl-mode\\n'
    elif [ "$COMP_WORDS" = 'mycli --completions ba' ]; then
        printf 'plain,bash\\n'
    fi
    exit 0
fi
if [ "$1" = '--list-dsn' ]; then
    printf 'list-dsn\\n' >> "$MYCLI_LOG"
    printf 'prod\\nstaging\\n'
fi
''',
        encoding='utf-8',
    )
    executable.chmod(executable.stat().st_mode | stat.S_IXUSR)
    (tmp_path / 'mysql.sock').touch()
    (tmp_path / 'space socket.sock').touch()
    (tmp_path / 'batch.sql').touch()
    log_path = tmp_path / 'calls.log'
    environment = {
        **os.environ,
        'COMPLETION_SCRIPT': str(COMPLETION_SCRIPT),
        'MYCLI_LOG': str(log_path),
        'PATH': f'{tmp_path}{os.pathsep}{os.environ.get("PATH", "")}',
    }

    result = subprocess.run(
        [BASH, '--noprofile', '--norc', '-c', BASH_COMPLETION_HARNESS],
        check=True,
        capture_output=True,
        text=True,
        cwd=tmp_path,
        env=environment,
    )

    assert result.stdout.splitlines() == [
        'complete -F _mycli_completion mycli',
        'generic|reply<--socket><--ssl-mode>|compopt',
        'completions|reply<bash>|compopt',
        'bare|reply<prod>|compopt',
        'dsn-separate|reply<prod><staging>|compopt',
        'dsn-attached|reply<-dprod>|compopt',
        'database-separate|reply<staging>|compopt',
        'database-attached|reply<--database=prod>|compopt',
        'after-host|reply<prod>|compopt',
        'host-value|reply|compopt',
        'after-double-dash|reply<staging>|compopt',
        'after-positional|reply|compopt',
        'clustered-option|reply<prod>|compopt',
        'socket-separate|reply<mysql.sock>|compopt<-o filenames>',
        'socket-attached|reply<--socket=space socket.sock>|compopt<-o filenames>',
        'socket-short-attached|reply<-Smysql.sock>|compopt<-o filenames>',
        'checkpoint-separate|reply<batch.sql>|compopt<-o filenames>',
        'batch-attached|reply<--batch=batch.sql>|compopt<-o filenames>',
    ]
    assert log_path.read_text(encoding='utf-8').splitlines().count('list-dsn') == 8


BASH_COMPLETION_HARNESS = r'''
compopt() {
    compopt_calls+=("$*")
}

source "$COMPLETION_SCRIPT"
complete -p mycli

run_completion() {
    local label="$1"
    local reply
    local call
    shift
    COMP_WORDS=("$@")
    COMP_CWORD=$((${#COMP_WORDS[@]} - 1))
    compopt_calls=()
    _mycli_completion

    printf '%s|reply' "$label"
    for reply in "${COMPREPLY[@]}"; do
        printf '<%s>' "$reply"
    done
    printf '|compopt'
    for call in "${compopt_calls[@]}"; do
        printf '<%s>' "$call"
    done
    printf '\n'
}

run_completion generic mycli --s
run_completion completions mycli --completions ba
run_completion bare mycli pro
run_completion dsn-separate mycli -d ''
run_completion dsn-attached mycli -dpro
run_completion database-separate mycli --database sta
run_completion database-attached mycli --database=pro
run_completion after-host mycli --host db pro
run_completion host-value mycli --host pro
run_completion after-double-dash mycli -- sta
run_completion after-positional mycli prod --host db other
run_completion clustered-option mycli -vP 3306 pro
run_completion socket-separate mycli --socket mysql
run_completion socket-attached mycli '--socket=space'
run_completion socket-short-attached mycli -Smysql
run_completion checkpoint-separate mycli --checkpoint batch
run_completion batch-attached mycli --batch=bat
'''



from __future__ import annotations

import os
from pathlib import Path
import shutil
import stat
import subprocess

import pytest

ZSH = shutil.which('zsh')
COMPLETION_SCRIPT = Path(__file__).parents[2] / 'mycli' / 'resources' / 'completions' / 'zsh' / '_mycli'


@pytest.mark.skipif(ZSH is None, reason='zsh is not installed')
def test_zsh_completion_lists_dsn_aliases_in_supported_contexts(tmp_path: Path) -> None:
    assert ZSH is not None
    executable = tmp_path / 'mycli'
    executable.write_text(
        '''#!/bin/sh
if [ "${_MYCLI_COMPLETE:-}" = 'zsh_complete' ]; then
    printf 'click\\n' >> "$MYCLI_LOG"
    if [ "$COMP_WORDS" = 'mycli --completions ba' ]; then
        printf 'plain\\nbash\\n_\\n'
    fi
    exit 0
fi
if [ "$1" = '--list-dsn' ]; then
    printf 'list-dsn\\n' >> "$MYCLI_LOG"
    printf 'prod\\nstaging\\n'
fi
''',
        encoding='utf-8',
    )
    executable.chmod(executable.stat().st_mode | stat.S_IXUSR)
    log_path = tmp_path / 'calls.log'
    environment = {
        **os.environ,
        'COMPLETION_SCRIPT': str(COMPLETION_SCRIPT),
        'MYCLI_LOG': str(log_path),
        'PATH': f'{tmp_path}{os.pathsep}{os.environ["PATH"]}',
    }

    result = subprocess.run(
        [ZSH, '-fc', ZSH_COMPLETION_HARNESS],
        check=True,
        capture_output=True,
        text=True,
        env=environment,
    )

    assert result.stdout.splitlines() == [
        'prod,staging||0|',
        'prod||0|',
        'bash||1|',
        'prod,staging||0|',
        'prod,staging||0|',
        '-dprod|-P -d|0|',
        '--dsn=prod|-P --dsn=|0|',
        '||0|',
        '||0|',
        '||0|-f',
        '||0|-f',
        '|-P --socket=|0|-f',
        '|-P -S|0|-f',
        'prod,staging||0|',
        'prod,staging||0|',
        '-Dprod|-P -D|0|',
        '--database=prod|-P --database=|0|',
        'prod||0|',
        'staging||0|',
        'prod||0|',
        '||0|',
        '||0|',
        'prod||0|',
        '||0|',
    ]
    assert log_path.read_text(encoding='utf-8').splitlines().count('list-dsn') == 14


ZSH_COMPLETION_HARNESS = r'''
compdef() { :; }
_describe() { :; }
_path_files() {
    path_file_calls+=("$*")
}

typeset -a captured compset_calls path_file_calls
typeset PREFIX IPREFIX used_unambiguous
compadd() {
    local array_name=''
    local candidate

    while (( $# )); do
        case "$1" in
            -U)
                used_unambiguous=1
                ;;
            -a)
                array_name="$2"
                shift
                ;;
        esac
        shift
    done

    for candidate in "${(@P)array_name}"; do
        [[ "$candidate" == "$PREFIX"* ]] && captured+=("$IPREFIX$candidate")
    done
}
compset() {
    compset_calls+=("$*")
    IPREFIX="$2"
    PREFIX="${PREFIX#$IPREFIX}"
}

source "$COMPLETION_SCRIPT"

run_completion() {
    captured=()
    compset_calls=()
    path_file_calls=()
    words=("$@")
    CURRENT=$#
    PREFIX="${words[CURRENT]}"
    IPREFIX=''
    used_unambiguous=0
    _mycli_completion
    print -r -- "${(j:,:)captured}|${(j:,:)compset_calls}|$used_unambiguous|${(j:,:)path_file_calls}"
}

run_completion mycli ''
run_completion mycli pro
run_completion mycli --completions ba
run_completion mycli -d ''
run_completion mycli --dsn ''
run_completion mycli -dpro
run_completion mycli --dsn=pro
run_completion mycli --host ''
run_completion mycli --host
run_completion mycli --socket ''
run_completion mycli -S ''
run_completion mycli --socket=/tmp/mysql
run_completion mycli -S/tmp/mysql
run_completion mycli -D ''
run_completion mycli --database ''
run_completion mycli -Dpro
run_completion mycli --database=pro
run_completion mycli --host db pro
run_completion mycli --ssl-mode auto sta
run_completion mycli -- prod
run_completion mycli --host pro
run_completion mycli --ssl-mode sta
run_completion mycli -vP 3306 pro
run_completion mycli prod --host db other
'''



from __future__ import annotations

from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qsl, urlencode, urlsplit, urlunsplit

from mycli.config import str_to_bool
from mycli.constants import DEFAULT_CHARSET, DEFAULT_PROMPT, KNOWN_DSN_QUERY_PARAMS

if TYPE_CHECKING:
    from mycli.client import MyCli

DSN_SUBCOMMANDS = {'help', 'list', 'show', 'save', 'delete'}
INVALID_DSN_ALIAS_ERROR = 'Error: DSN aliases cannot start with a dash.'

SSL_QUERY_PARAMS = {
    'ssl_ca': 'ca',
    'ssl_capath': 'capath',
    'ssl_cert': 'cert',
    'ssl_cipher': 'cipher',
    'ssl_key': 'key',
    'ssl_mode': 'mode',
    'ssl_verify_server_cert': 'check_hostname',
    'tls_version': 'tls_version',
}


def is_valid_dsn_alias(alias: str) -> bool:
    return not alias.startswith('-')


def _config_bool(value: Any) -> bool:
    try:
        return str_to_bool(value)
    except (TypeError, ValueError):
        return False


def _query_value(value: Any) -> str:
    if isinstance(value, bool):
        return str(value).lower()
    return str(value)


class DsnAliases:
    section_name: str = 'alias_dsn'

    usage = """
DSN aliases are a way to save frequently used connections
with a short alias.  You can manage them by editing ~/.myclirc
directly or by using this command.

The only way to include a password in a DSN alias is by editing
~/.myclirc directly.

Examples:

    # Show a DSN for the current connection
    mysql> /dsn show
    mysql://mycli@localhost/mysql

    # Include non-default connection settings
    mysql> /dsn show --more
    mysql://mycli@localhost/mysql?prompt=%5Cd%3E%5C_

    # Save a DSN alias based on the current connection
    # The password will not be saved!
    mysql> /dsn save rocks

    # Include non-default connection settings in the saved DSN
    # The password will still not be saved!
    mysql> /dsn save --more rocks

    # List all persisted DSN aliases, some of which might have been
    # saved with --more
    mysql> /dsn list
    ┌───────┬──────────────────────────────────────────────────┐
    │ Alias │ DSN                                              │
    ├───────┼──────────────────────────────────────────────────┤
    │ rocks │ mysql://mycli@localhost/mysql?prompt=%5Cd%3E%5C_ │
    └───────┴──────────────────────────────────────────────────┘

    # Delete a DSN alias.
    mysql> /dsn delete rocks
"""

    # Class-level variable, for convenience to use as a singleton.
    instance: DsnAliases

    def __init__(self, config: Any, mycli: MyCli | None = None) -> None:
        self.config = config
        self.mycli = mycli

    @classmethod
    def from_config(cls, config: Any, mycli: MyCli | None = None) -> DsnAliases:
        return DsnAliases(config, mycli)

    def _query_param_defaults(self) -> dict[str, Any]:
        if self.mycli is None:
            return {}

        main_config = self.config.get('main', {})
        connection_config = self.config.get('connection', {})
        vault_config = self.config.get('vault_beta', {})
        return {
            'character_set': connection_config.get('default_character_set') or DEFAULT_CHARSET,
            'keepalive_ticks': self.mycli.default_keepalive_ticks,
            'prompt': main_config.get('prompt') or DEFAULT_PROMPT,
            'socket': connection_config.get('default_socket') or '',
            'ssl_ca': connection_config.get('default_ssl_ca') or '',
            'ssl_capath': connection_config.get('default_ssl_capath') or connection_config.get('default_ssl_ca_path') or '',
            'ssl_cert': connection_config.get('default_ssl_cert') or '',
            'ssl_cipher': connection_config.get('default_ssl_cipher') or '',
            'ssl_key': connection_config.get('default_ssl_key') or '',
            'ssl_mode': self.mycli.ssl_mode,
            'ssl_verify_server_cert': _config_bool(connection_config.get('default_ssl_verify_server_cert')),
            'vault_address': vault_config.get('address') or '',
            'vault_mount': vault_config.get('default_mount') or '',
            'vault_password_field': vault_config.get('default_password_field') or 'password',
            'vault_username_field': vault_config.get('default_username_field') or 'username',
        }

    def dsn_more(self, dsn: str) -> str:
        if self.mycli is None or self.mycli.sqlexecute is None:
            return dsn

        parsed = urlsplit(dsn)
        query_params: dict[str, Any] = {
            key: value for key, value in parse_qsl(parsed.query, keep_blank_values=True) if key in KNOWN_DSN_QUERY_PARAMS
        }
        sqlexecute = self.mycli.sqlexecute
        query_params.update({
            'character_set': sqlexecute.character_set,
            'keepalive_ticks': self.mycli.keepalive_ticks,
            'prompt': self.mycli.prompt_format,
        })
        ssl = sqlexecute.ssl or {}
        query_params.update({query_param: ssl.get(ssl_key) for query_param, ssl_key in SSL_QUERY_PARAMS.items()})

        defaults = self._query_param_defaults()
        more_params = [
            (key, _query_value(query_params[key]))
            for key in sorted(KNOWN_DSN_QUERY_PARAMS)
            if query_params.get(key) not in (None, '', False) and _query_value(query_params[key]) != _query_value(defaults.get(key, ''))
        ]
        return urlunsplit(parsed._replace(query=urlencode(more_params)))

    def list(self) -> list[str]:
        return [alias for alias in self.config.get(self.section_name, {}) if is_valid_dsn_alias(alias)]

    def get(self, alias: str) -> str | None:
        if not is_valid_dsn_alias(alias):
            return None
        return self.config.get(self.section_name, {}).get(alias, None)

    def save(self, alias: str, dsn: str) -> str:
        if not is_valid_dsn_alias(alias):
            return INVALID_DSN_ALIAS_ERROR
        self.config.encoding = 'utf-8'
        if self.section_name not in self.config:
            self.config[self.section_name] = {}
        self.config[self.section_name][alias] = dsn
        self.config.write()
        return f'Saved: {alias}'

    def delete(self, alias: str) -> str:
        if not is_valid_dsn_alias(alias):
            return INVALID_DSN_ALIAS_ERROR
        try:
            del self.config[self.section_name][alias]
        except KeyError:
            return f'Not Found: {alias}'
        self.config.write()
        return f'Deleted: {alias}'
